unit UTileMap;

interface

uses
   Windows, USmUtil;

const
	MAX_MAP_EXIT_POINT = 10;
type

   TMapAttribute = ( maTile, maHeight, maPerson, maWall1, maObject, maAttribute, maAux );

   TEventMessage = ( emNone, emInEvent, emOnEvent, emHitEvent, emSearchEvent, emWalkEvent, emBlockedEvent, emDarkEvent, emTalkEvent, emTimeEvent, emDramaMode );

   TMapFillMode  = ( mfBoundary, mfForced, mfDecHeight );

   TMapHeaderData = record
	  FileName : string[8];
	  MapStart : TPoint;
	  MapExit  : TPoint;
   end;

   TMapHeader     = record
	  ID              : string[6];
	  MapCode         : string[4];
	  MapData         : TMapHeaderData;
	  BoundaryExit    : boolean;
	  HandicapBit     : SmallInt;
	  NumExitPoint    : byte;
	  FriendshipLevel : byte;
	  MapRecording    : boolean;
	  PersonRecording : boolean;
	  MapExitData     : array[1..MAX_MAP_EXIT_POINT] of TMapHeaderData;
	  MapSize         : TPoint;
	  MIDI            : string[8];
	  BMP             : string[8];
	  GABAGE          : array[1..50-18] of byte;
   end;

   PMapField     = ^TMapField;
   TMapField     = packed record
	  Tile               : SmallInt;
	  Wall               : array[1..1] of SmallInt;
	  Objects, Attribute : byte;
	  Height, Person     : byte;
   end;

   TTileMap = class

	  constructor Create;
	  procedure   Free;

   private
	  FMapData    : PMapField;
	  FMapClip    : PMapField;
	  FMapAux     : PByte;        // ð  ϴ  
	  FClipSize   : TPoint;
	  FCurrentMap : string[16];

	  procedure   Initialze;
	  procedure   Finalize;

	  function    ReadMap ( attribute : TMapAttribute; x, y : integer ) : SmallInt;
	  procedure   WriteMap( attribute : TMapAttribute; x, y : integer; value : SmallInt );
	  function    ReadEventMap ( EventMessage : TEventMessage; x, y : integer ) : boolean;
	  procedure   WriteEventMap( EventMessage : TEventMessage; x, y : integer; Value : boolean );

   public

	  MapHeader      : TMapHeader;
	  CurrentMapAxis : TPoint;

	  function    IsMovableMap( _x, _y, x, y : integer ) : boolean;
	  function    IsStandableMap( _x, _y, x, y : integer ) : boolean;
	  property    Map[attribute : TMapAttribute; x, y : integer]         : SmallInt read ReadMap      write WriteMap; default;
	  property    EventMap[EventMessage : TEventMessage; x, y : integer] : boolean  read ReadEventMap write WriteEventMap;

	  procedure   CopyMap ( ARect  : TRect );
	  procedure   PasteMap( APoint : TPoint );
	  procedure   FillMap ( ARect  : TRect; tile, height : integer; FillMode : TMapFillMode );
	  procedure   InitializeMap( AMapHeader : TMapHeader );
	  function    LoadMap( LoadHeader : boolean ) : boolean;
	  function    LoadMapWithHeader( AMapHeader : TMapHeader ) : boolean;
	  function    LoadMapWithName( Name : string ) : boolean;
	  procedure   SaveMap;

	  procedure   LoadMapFromFile( var f : file );
	  procedure   SaveMapToFile( var f : file );
	  procedure   RandomMap( Width, Height : Integer );
   end;

implementation

constructor TTileMap.Create;
begin
	inherited;
	Initialze;
end;

procedure   TTileMap.Free;
begin
	Finalize;
	inherited;
end;

procedure   TTileMap.Initialze;
begin
	FMapData    := nil;
	FillChar( MapHeader, sizeof( TMapHeader ), 0 );

	FMapClip    := nil;
	FClipSize.x := 0;
	FClipSize.y := 0;

	FMapAux     := nil;

	CurrentMapAxis.x := 0;
	CurrentMapAxis.y := 0;
end;

procedure   TTileMap.Finalize;
begin
	if Assigned( FMapAux ) then begin
		FreeMem( FMapAux );
		FMapAux  := nil;
	end;

	if Assigned( FMapClip ) then begin
		FreeMem( FMapClip );
		FMapClip := nil;
	end;

	if Assigned( FMapData ) then begin
		FreeMem( FMapData );
		FMapData := nil;
	end;
end;

function    TTileMap.ReadEventMap(EventMessage : TEventMessage; x, y : integer) : boolean;
var
	Attribute : byte;
	Key       : byte;
begin
	case EventMessage of
		emInEvent      : Key := 1;
		emOnEvent      : Key := 2;
		emHitEvent     : Key := 3;
		emSearchEvent  : Key := 4;
		emWalkEvent    : Key := 5;
		emBlockedEvent : Key := 6;
		emDarkEvent    : Key := 7;
		else begin ReadEventMap := FALSE; Exit end;
	end;

	Attribute := Map[maAttribute,x,y];

	if Key in [1..3] then begin
		ReadEventMap := ( Attribute and Byte( $03 ) ) = Key;
	end else begin
		ReadEventMap := ( Attribute and Byte( 1 shl ( Key - 2 ) ) ) > 0;
	end;
end;

procedure   TTileMap.WriteEventMap( EventMessage : TEventMessage; x, y : integer; Value : boolean );
var
	Attribute : byte;
	Key       : byte;
begin
	case EventMessage of
		emInEvent      : Key := 1;
		emOnEvent      : Key := 2;
		emHitEvent     : Key := 3;
		emSearchEvent  : Key := 4;
		emWalkEvent    : Key := 5;
		emBlockedEvent : Key := 6;
		emDarkEvent    : Key := 7;
		emNone         : begin Map[maAttribute,x,y] := 0; exit end;
		else             Exit;
	end;

	Attribute := Map[maAttribute,x,y];

	if Key in [1..3] then begin
		if Value = FALSE then Key := 0;
		Attribute := ( Attribute and Byte( $FC ) ) or Key;
	end else begin
		if Value = TRUE  then begin
			Attribute := Attribute or Byte( 1 shl ( Key - 2 ) );
		end else begin
			Attribute := Attribute and ( not Byte( 1 shl ( Key - 2 ) ) );
		end;
	end;

	Map[maAttribute,x,y] := Attribute;
end;

function    TTileMap.ReadMap( attribute : TMapAttribute; x, y : integer ) : SmallInt;

 function  isUnvalancedMap( x, y : integer; height : byte ) : boolean;
 begin
	isUnvalancedMap := FALSE;
	if ( x >= 0 ) and ( y >= 0 ) and ( x < MapHeader.MapSize.X ) and ( y < MapHeader.MapSize.Y ) then begin
	   if ReadMap( maHeight, x, y ) < height then isUnvalancedMap := TRUE;
	end;
 end;

var
   tile_no : SmallInt;
   pb      : PMapField;
   pAux    : PByte;
begin
	if ( x >= 0 ) and ( y >= 0 ) and ( x < MapHeader.MapSize.X ) and ( y < MapHeader.MapSize.Y ) then begin
		pb := FMapData;
		Inc( DWORD( pb ), ( MapHeader.MapSize.X * y + x ) * sizeof( TMapField ) );

		if attribute = maTile then begin
			tile_no := pb^.Tile * 4;
			if IsUnvalancedMap( x, y-1, pb^.Height ) then Inc( tile_no, 1 );
			if IsUnvalancedMap( x+1, y, pb^.Height ) then Inc( tile_no, 2 );

		end else if attribute = maHeight    then begin
			tile_no := pb^.Height;
		end else if attribute = maPerson    then begin
			tile_no := pb^.Person;
		end else if attribute = maWall1     then begin
			tile_no := pb^.Wall[1];
		end else if attribute = maObject    then begin
			tile_no := pb^.Objects;
		end else if attribute = maAttribute then begin
			tile_no := pb^.Attribute;
		end else if attribute = maAux       then begin
			pAux := FMapAux;
			Inc( DWORD( pAux ), ( MapHeader.MapSize.X * y + x ) * sizeof( Byte ) );
			tile_no := pAux^;
		end else tile_no := 0;
	end else begin
		if attribute = maTile then tile_no := 4 * 4
							  else tile_no := 0
	end;

	ReadMap := tile_no;
end;

procedure   TTileMap.WriteMap( attribute : TMapAttribute; x, y : integer; value : SmallInt );
var
	pb      : PMapField;
	pAux    : PByte;
begin
	if ( x >= 0 ) and ( y >= 0 ) and ( x < MapHeader.MapSize.X ) and ( y < MapHeader.MapSize.Y ) then begin
		pb := FMapData;
		Inc( DWORD( pb ), ( MapHeader.MapSize.X * y + x ) * sizeof( TMapField ) );

		if attribute = maHeight then begin
			if shortint( value ) <  0 then value := 0;
			if shortint( value ) > 10 then value := 10;
		end;
		if attribute = maTile then begin
			pb^.Tile      := value;
		end else if attribute = maHeight    then begin
			pb^.Height    := value;
		end else if attribute = maPerson    then begin
			pb^.Person    := value;
		end else if attribute = maWall1     then begin
			pb^.Wall[1]   := value;
		end else if attribute = maObject    then begin
			pb^.Objects   := value;
		end else if attribute = maAttribute then begin
			pb^.Attribute := value;
		end else if attribute = maAux       then begin
			pAux := FMapAux;
			Inc( DWORD( pAux ), ( MapHeader.MapSize.X * y + x ) * sizeof( Byte ) );
			pAux^         := Byte( value );
		end;
	end;
end;

function    TTileMap.IsMovableMap(_x, _y, x, y : integer) : boolean;
begin
	IsMovableMap := FALSE;

	if ( x >= 0 ) and ( y >= 0 ) and ( x < MapHeader.MapSize.X ) and ( y < MapHeader.MapSize.Y ) then begin
		IsMovableMap := TRUE;
	end;
end;

function   TTileMap.IsStandableMap(_x, _y, x, y : integer) : boolean;
var
	Tile : Integer;
begin
	IsStandableMap := FALSE;

	if (x >= 0) and (y >= 0) and (x < MapHeader.MapSize.X) and (y < MapHeader.MapSize.Y) then begin
		Tile := Map[maTile,x,y];

		// Walkable Event  쿡    ִ.
		if EventMap[emWalkEvent,x,y] then begin
			IsStandableMap := TRUE;
			Exit;
		end;

		// Blocked Event  쿡    .
		if EventMap[emBlockedEvent,x,y] then
			Exit;

		if ( Map[maHeight, _x, _y] + 1 >= Map[maHeight, x, y] ) then
			IsStandableMap := TRUE;
   end;
end;

procedure   TTileMap.CopyMap(ARect : TRect);
var
	i, x, y      : integer;
	pSour, pDest : PMapField;
begin
	if Assigned( FMapClip ) then begin
		FreeMem( FMapClip );
		FMapClip := nil;
	end;

	if ARect.Right < ARect.Left then begin
		i := ARect.Right;
		ARect.Right := ARect.Left;
		ARect.Left  := i;
	end;
	if ARect.Bottom < ARect.Top then begin
		i := ARect.Bottom;
		ARect.Bottom := ARect.Top;
		ARect.Top    := i;
	end;

	if ARect.Left   <  0                   then ARect.Left   := 0;
	if ARect.Top    <  0                   then ARect.Top    := 0;
	if ARect.Right  >= MapHeader.MapSize.X then ARect.Right  := Pred( MapHeader.MapSize.X );
	if ARect.Bottom >= MapHeader.MapSize.Y then ARect.Bottom := Pred( MapHeader.MapSize.Y );

	FClipSize.x := Succ( ARect.Right  - ARect.Left );
	FClipSize.y := Succ( ARect.Bottom - ARect.Top  );

	GetMem( FMapClip, sizeof( TMapField ) * FClipSize.X * FClipSize.Y );

	pDest := FMapClip;

	for y := ARect.Top to ARect.Bottom do begin
		pSour := FMapData;
		Inc( DWORD( pSour ), ( MapHeader.MapSize.X * y + ARect.Left ) * sizeof( TMapField ) );
		for x := ARect.Left to ARect.Right do begin
			pDest^ := pSour^;
			Inc( DWORD( pSour ), sizeof( TMapField ) );
			Inc( DWORD( pDest ), sizeof( TMapField ) );
		end;
	end;
end;

procedure   TTileMap.PasteMap(APoint : TPoint);
var
	x, y         : integer;
	pSour, pDest : PMapField;
begin
	if Assigned( FMapClip ) then begin
		pSour := FMapClip;
		for y := APoint.Y to APoint.Y + Pred( FClipSize.Y ) do begin
			pDest := FMapData;
			Inc( DWORD( pDest ), ( MapHeader.MapSize.X * y + APoint.X ) * sizeof( TMapField ) );
			for x := APoint.X to APoint.X + Pred( FClipSize.X ) do begin
				if ( x >= 0 ) and ( y >= 0 ) and ( x < MapHeader.MapSize.X ) and ( y < MapHeader.MapSize.Y ) then begin
					pDest^ := pSour^;
				end;
				Inc( DWORD( pSour ), sizeof( TMapField ) );
				Inc( DWORD( pDest ), sizeof( TMapField ) );
			end;
		end;
	end;
end;

procedure   TTileMap.FillMap( ARect : TRect; tile, height : integer; FillMode : TMapFillMode );
var
	i, j      : integer;
	FixedTile : integer;

 procedure FillMapSub( x, y : integer; tile, height : integer );
 begin
	if ( x >= 0 ) and ( y >= 0 ) and ( x < MapHeader.MapSize.X ) and ( y < MapHeader.MapSize.Y ) then begin
	   if Map[maTile,x,y] = FixedTile then begin
		  if tile >= 0 then
			 Map[maTile  ,x,y] := tile;
		  if height >= 0 then
			 Map[maHeight,x,y] := height;
	   end;

	   if ( Map[maTile,x-1,y] = FixedTile ) then FillMapSub( x-1, y, tile,height );
	   if ( Map[maTile,x+1,y] = FixedTile ) then FillMapSub( x+1, y, tile,height );
	   if ( Map[maTile,x,y-1] = FixedTile ) then FillMapSub( x, y-1, tile,height );
	   if ( Map[maTile,x,y+1] = FixedTile ) then FillMapSub( x, y+1, tile,height );
	end;
 end;

 procedure DecHeightSub( _x, _y : integer; _height, level : integer );
 begin
	if ( ( abs( ARect.Left-_x ) + abs( ARect.Top-_y ) ) = level ) and ( _height >= 0 ) then
	if ( _x >= 0 ) and ( _y >= 0 ) and ( _x < MapHeader.MapSize.X ) and ( _y < MapHeader.MapSize.Y ) then begin
	   Map[maHeight,_x,_y] := _height;
	   DecHeightSub( _x-1, _y, _height-1, level + 1 );
	   DecHeightSub( _x+1, _y, _height-1, level + 1 );
	   DecHeightSub( _x, _y-1, _height-1, level + 1 );
	   DecHeightSub( _x, _y+1, _height-1, level + 1 );
	end;
 end;

begin
	case FillMode of
		mfBoundary  :
		begin
			if ( ARect.Left >= 0 ) and ( ARect.Top >= 0 ) and ( ARect.Left < MapHeader.MapSize.X ) and ( ARect.Top < MapHeader.MapSize.Y ) then begin
				FixedTile := Map[maTile  ,ARect.Left,ARect.Top];
				if FixedTile <> tile then
					FillMapSub( ARect.Left, ARect.Top, tile,height );
			end;
		end;
		mfForced    :
		for j := ARect.Top to ARect.Bottom do begin
			for i := ARect.Left to ARect.Right do begin
				if tile >= 0 then
					Map[maTile  ,i,j] := tile;
				if height >= 0 then
					Map[maHeight,i,j] := height;
			end;
		end;
		mfDecHeight : DecHeightSub( ARect.Left, ARect.Top, height, 0 );
	end;
end;

procedure   TTileMap.InitializeMap( AMapHeader : TMapHeader );
begin
	if Assigned( FMapData ) then begin
		FreeMem( FMapData );
		FMapData := nil;
	end;

	MapHeader := AMapHeader;

	FCurrentMap := MapHeader.MapData.FileName;
	if Pos( '.', FCurrentMap ) = 0 then FCurrentMap := FCurrentMap + '.HMp';

	GetMem( FMapData, sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
	FillChar( FMapData^, sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y, 0 );
	FillMap( Rect( 0, 0, Pred( MapHeader.MapSize.X ),Pred( MapHeader.MapSize.Y ) ), 0, 0, mfForced );

	GetMem( FMapAux, sizeof( Byte ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
	FillChar( FMapAux^, sizeof( Byte ) * MapHeader.MapSize.X * MapHeader.MapSize.Y, 1 );
end;

function    TTileMap.LoadMap( LoadHeader : boolean ) : boolean;
var
	f          : File;
	i, j       : integer;
	TempHeader : TMapHeader;
begin
	LoadMap    := FALSE;

	AssignFile( f, FCurrentMap );
 {$I-}
	Reset( f, 1 );
 {$I+}

	if Assigned( FMapData ) then begin
		FreeMem( FMapData );
		FMapData := nil;
	end;

	if IOResult = 0 then begin

		if LoadHeader then
			BlockRead( f, MapHeader,  sizeof( MapHeader ) )
		else
			BlockRead( f, TempHeader, sizeof( TempHeader ) );

		GetMem( FMapData, sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
		BlockRead( f, FMapData^, sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );

		GetMem( FMapAux, sizeof( Byte ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
		FillChar( FMapAux^, sizeof( Byte ) * MapHeader.MapSize.X * MapHeader.MapSize.Y, 1 );

		CloseFile( f );
{
		for k := 1 to AVAILABLE_PERSON do
		for j := 0 to pred( MapHeader.MapSize.Y ) do begin
			for i := 0 to pred( MapHeader.MapSize.X ) do begin
				if Map[maPerson,i,j] = GameSystem.person[k].mNumber then begin
					GameSystem.person[k].mX := i;
					GameSystem.person[k].mY := j;
				end;
			end;
		end;
}
		for j := 0 to pred( MapHeader.MapSize.Y ) do
		for i := 0 to pred( MapHeader.MapSize.X ) do begin
			Map[maPerson,i,j] := 0;
		end;

		LoadMap := TRUE;
	end;

end;

function    TTileMap.LoadMapWithHeader( AMapHeader : TMapHeader ) : boolean;
var
	Name : string;
begin
	Name := AMapHeader.MapData.FileName;
	if Pos( '.', Name ) = 0 then Name := Name + '.HMp';

	if FileExists( Name ) then begin
		FCurrentMap       := Name;
		MapHeader         := AMapHeader;
		LoadMapWithHeader := LoadMap( FALSE );
	end else begin
		LoadMapWithHeader := FALSE;
	end;
end;

function    TTileMap.LoadMapWithName( Name : string ) : boolean;
begin
	if Pos( '.', Name ) = 0 then Name := Name + '.HMp';

	if FileExists( Name ) then begin
		FCurrentMap     := Name;
		LoadMapWithName := LoadMap( TRUE );
	end else begin
		LoadMapWithName := FALSE;
	end;
end;

procedure   TTileMap.SaveMap;
var
	f : File;
begin
	AssignFile( f, FCurrentMap );
 {$I-}
	Rewrite( f, 1 );
 {$I+}
	if IOResult <> 0 then exit;
	BlockWrite( f, MapHeader, sizeof( MapHeader ) );
	BlockWrite( f, FMapData^, sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
	CloseFile( f );
end;

procedure   TTileMap.LoadMapFromFile( var f : file );
begin
   Finalize;
   Initialze;
   BlockRead( f, MapHeader,   sizeof( MapHeader ) );
   BlockRead( f, FCurrentMap, sizeof( FCurrentMap ) );
   GetMem   (    FMapData ,   sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
   BlockRead( f, FMapData^,   sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
   GetMem   (    FMapAux  ,   sizeof( Byte )      * MapHeader.MapSize.X * MapHeader.MapSize.Y );
   BlockRead( f, FMapAux^ ,   sizeof( Byte )      * MapHeader.MapSize.X * MapHeader.MapSize.Y );
end;

procedure   TTileMap.SaveMapToFile( var f : file );
begin
	BlockWrite( f, MapHeader, sizeof( MapHeader ) );
	BlockWrite( f, FCurrentMap, sizeof( FCurrentMap ) );
	BlockWrite( f, FMapData^, sizeof( TMapField ) * MapHeader.MapSize.X * MapHeader.MapSize.Y );
	BlockWrite( f, FMapAux^ , sizeof( Byte )      * MapHeader.MapSize.X * MapHeader.MapSize.Y );
end;

procedure   TTileMap.RandomMap( Width, Height : Integer );
var
	MapHeader : TMapHeader;
	x, y      : Integer;
begin
	FillChar( MapHeader, sizeof( MapHeader ), 0 );

	MapHeader.MapSize.x := Width;
	MapHeader.MapSize.y := Height;

	InitializeMap( MapHeader );

	for y := 0 to Pred( Height ) do begin
		for x := 0 to Pred( Width  ) do begin
			WriteMap( maTile, x, y, Random( 4 ) );
		end;
	end;
	WriteMap( maTile, 0, 0, 5 );
end;

end.
